local HeadCount = LibStub("AceAddon-3.0"):GetAddon("HeadCount2")
local L = LibStub("AceLocale-3.0"):GetLocale("HeadCount2")

local Fields = {
	playerList = {},
	bossList = {},
	lootList = {},
	timeList = {},
	difficulty = 0,
	raidTime = 0,
	lastActivityTime = nil,
	zone = UKNOWN,
	numberOfPlayers = 0,
	numberOfBosses = 0,
	isFinalized = false,
	lastBoss = nil
}

local Raid = HeadCount:NewClassFunctions()

function Raid:Init(args)
	if args then
		self.playerList = args["playerList"]
		self.bossList = args["bossList"]
		self.lootList = args["lootList"]
		self.timeList = args["timeList"]
		self.difficulty = args["difficulty"]
		self.raidTime = args["raidTime"]
		self.lastActivityTime = args["lastActivityTime"]
		self.zone = args["zone"]
		self.numberOfPlayers = args["numberOfPlayers"]
		self.numberOfBosses = args["numberOfBosses"]
		self.isFinalized = args["isFinalized"]
		self.timeList = args["timeList"] or {}
		self.lastBoss = args["lastBoss"] or HeadCount.DEFAULT_LOOT_SOURCE
	end
end

-- Adds a boss to the boss list
-- @param boss The boss to add to the boss list
function Raid:addBoss(boss)
	if boss then
		local name = boss.name
		if not self.bossList[name] then
			-- boss is not present, add it
			self.bossList[name] = boss
			self.numberOfBosses = self.numberOfBosses + 1
			self.lastBoss = name
		end
	end
end

-- Removes a boss from the boss list.
-- @param bossName The boss name to remove from the boss list.
function Raid:removeBoss(bossName)
	if self.bossList[bossName] then
		self.bossList[bossName] = nil
		self.numberOfBosses = self.numberOfBosses - 1
	end
end

-- Removes an event attendee from a boss kill.
-- @param bossName The boss name.
-- @param attendeeId The attendee id to remove.
function Raid:removeEventAttendee(bossName, attendeeId)
	if self.bossList[bossName] then
		local boss = self.bossList[bossName]
		boss:removePlayer(attendeeId)
	end
end

-- Retrieves a boss from the boss list.
-- @param bossName The boss name.
-- @return object Returns a boss from the boss list or nil if the player does not exist.
function Raid:retrieveBoss(bossName)
	return self.bossList[bossName]
end

function Raid:GetNumberOfPlayers()
	return self.numberOfPlayers
end

-- Adds a player to the player list
-- @param player The player to add to the player list
function Raid:addPlayer(player)
	if player then
		-- player exists
		local name = player.name
		if not self.playerList["raidgroup"][name] and not self.playerList["nogroup"][name] then
			-- player is not in the player list
			local isPlayerInRaidList = player.isPresentInRaidList
			local isPlayerInWaitList = player.isPresentInWaitList
			if isPlayerInRaidList or isPlayerInWaitList then
				-- player is currently in the raid
				self.playerList["raidgroup"][name] = player
			else
				-- player is currently NOT in the raid
				self.playerList["nogroup"][name] = player
			end
			self.numberOfPlayers = self.numberOfPlayers + 1
		end
	end
end

-- Removes a player from the player list
-- @param playerName The player name to remove from the player list
function Raid:removePlayer(playerName)
	local isPresent, groupName = self:isPlayerPresent(playerName)
	if isPresent then
		-- player is being tracked, remove them
		self.playerList[groupName][playerName] = nil
		self.numberOfPlayers = self.numberOfPlayers - 1
	end
end

-- Removes a player's wait list status.
-- @param playerName The player name.
function Raid:removeWaitlistPlayer(playerName)
	local isPresent, groupName = self:isPlayerPresent(playerName)
	if isPresent then
		-- player is being tracked, remove them
		self.playerList[groupName][playerName].isWaitlisted = false
		self.playerList[groupName][playerName].waitlistActivityTime = nil
		self.playerList[groupName][playerName].waitlistNote = nil
	end
end

-- Retrieves an ordered boss list.
-- @return table Returns an ordered boss list.
function Raid:retrieveOrderedBossList()
	-- Sort the player list by the given criteria
	local orderedBossList = {}
	for k,v in pairs(self.bossList) do
		table.insert(orderedBossList, v)
	end
	table.sort(orderedBossList, function(a, b)
		local aActivityTime = a.runtime
		local bActivityTime = b.runtime
		if aActivityTime and aActivityTime then
			-- both a and b have valid activity times
			local aSeconds = aActivityTime.epoc
			local bSeconds = bActivityTime.epoc
			return aSeconds < bSeconds
		end
	end)

	return orderedBossList
end

-- Determines if a player is present in the player list
-- @param playerName The player name to check for in the player list.
-- @return boolean Returns true if the player is present, false otherwise.
function Raid:isPlayerPresent(playerName)
	local isPresent = false
	local groupName = nil
	if self.playerList["raidgroup"][playerName] then
		-- player is present in the raid group
		isPresent = true
		groupName = "raidgroup"
	elseif self.playerList["nogroup"][playerName] then
		-- player is present in the missing group
		isPresent = true
		groupName = "nogroup"
	end
	return isPresent, groupName
end

-- Retrieves a player from the player list.
-- @param playerName The player name key.
-- @return object Returns a player from the player list or nil if the player does not exist.
function Raid:retrievePlayer(playerName)
	local player = nil
	local isPresent, groupName = self:isPlayerPresent(playerName)
	if isPresent then
		-- player exists
		player = self.playerList[groupName][playerName]
	end
	return player
end

-- Retrieves the player list by class
-- @return table Returns a player list table.
function Raid:retrievePlayerListByClass()
	local playerListByClass = {
		["DEATHKNIGHT"] = {},
		["DRUID"] = {},
		["HUNTER"] = {},
		["MAGE"] = {},
		["PALADIN"] = {},
		["PRIEST"] = {},
		["ROGUE"] = {},
		["SHAMAN"] = {},
		["WARLOCK"] = {},
		["WARRIOR"] = {},
		["UNKNOWN"] = {},
	}

	for k,v in pairs(self.playerList) do
		for i,j in pairs(self.playerList[k]) do
			local playerName = j.name
			local fileName = j.fileName
			if "DEATHKNIGHT" == fileName then
				table.insert(playerListByClass["DEATHKNIGHT"], playerName)
			elseif "DRUID" == fileName then
				table.insert(playerListByClass["DRUID"], playerName)
			elseif "HUNTER" == fileName then
				table.insert(playerListByClass["HUNTER"], playerName)
			elseif "MAGE" == fileName then
				table.insert(playerListByClass["MAGE"], playerName)
			elseif "PALADIN" == fileName then
				table.insert(playerListByClass["PALADIN"], playerName)
			elseif "PRIEST" == fileName then
				table.insert(playerListByClass["PRIEST"], playerName)
			elseif "ROGUE" == fileName then
				table.insert(playerListByClass["ROGUE"], playerName)
			elseif "SHAMAN" == fileName then
				table.insert(playerListByClass["SHAMAN"], playerName)
			elseif "WARLOCK" == fileName then
				table.insert(playerListByClass["WARLOCK"], playerName)
			elseif "WARRIOR" == fileName then
				table.insert(playerListByClass["WARRIOR"], playerName)
			else
				table.insert(playerListByClass["UNKNOWN"], playerName)
			end
		end
	end

	return playerListByClass
end

-- Retrieves an ordered player list.
-- @param sortType The sort type.
-- @param isRaidListGroupEnabled The raid list group enabled status.
-- @param isWaitListGroupEnabled The wait list group enabled status.
-- @param isNoListGroupEnabled The no list group enabled status.
-- @param The waitlisted flag The standby wait list status.
-- @param isNameOnly The name only flag.
-- @param isDescending The sort order.
-- @return table Returns an ordered player list.
function Raid:retrieveOrderedPlayerList(sortType, isRaidListGroupEnabled, isWaitListGroupEnabled, isNoListGroupEnabled, isWaitlisted, isNameOnly, isDescending)
	-- Sort the player list by the given criteria
	local orderedPlayerList = {}

	for k,v in pairs(self.playerList) do
		for i,j in pairs(self.playerList[k]) do
			if isRaidListGroupEnabled and j.isPresentInRaidList then
				--  raid list group enabled and player is in a raid list group, include them
				if isNameOnly then
					table.insert(orderedPlayerList, j.name)
				else
					table.insert(orderedPlayerList, j)
				end
			elseif isWaitListGroupEnabled and j.isPresentInWaitList then
				-- wait list group enabled and player is in a wait list group, include them
				if isNameOnly then
					table.insert(orderedPlayerList, j.name)
				else
					table.insert(orderedPlayerList, j)
				end
			elseif isNoListGroupEnabled and not (j.isPresentInRaidList or j.isPresentInWaitList) then
				-- no list group enabled and player is in a no list group, include them
				if isNameOnly then
					table.insert(orderedPlayerList, j.name)
				else
					table.insert(orderedPlayerList, j)
				end
			elseif isWaitlisted and j.isWaitlisted then
				-- wait list is enabled and player is on wait list, include them
				if isNameOnly then
					table.insert(orderedPlayerList, j.name)
				else
					table.insert(orderedPlayerList, j)
				end
			end
		end
	end

	if isNameOnly then
		table.sort(orderedPlayerList, function(a, b)
			return a < b
		end)
	else
		local raidMemberSort = HeadCount:getRAID_MEMBER_SORT()
		if raidMemberSort[sortType] then
			-- sort type is valid
			if raidMemberSort["Name"] == sortType then
				self:sortPlayerListByName(orderedPlayerList, isDescending)
			elseif raidMemberSort["Start"] == sortType then
				self:sortPlayerListByStartTime(orderedPlayerList, isDescending)
			elseif raidMemberSort["End"] == sortType then
				self:sortPlayerListByEndTime(orderedPlayerList, isDescending)
			elseif raidMemberSort["Waitlist"] == sortType then
				self:sortPlayerListByWaitlistActivityTime(orderedPlayerList, isDescending)
			elseif raidMemberSort["WaitlistNote"] == sortType then
				self:sortPlayerListByWaitlistNote(orderedPlayerList, isDescending)
			elseif raidMemberSort["Total"] == sortType then
				if self.isFinalized then
					-- raid is finalized, sort by actual total time
					self:sortPlayerListByTotalTime(orderedPlayerList, isDescending, isFinalized)
				else
					-- raid is not finalized, sort by name
					self:sortPlayerListByName(orderedPlayerList, isDescending)
				end
			end
		end
	end

	return orderedPlayerList
end

-- Sorts a player list by name
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by name
function Raid:sortPlayerListByName(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aName = a.name
			local bName = b.name

			if aName and bName then
				-- both a and b have valid names
				if isDescending then
					return aName < bName
				else
					return aName > bName
				end
			else
				--HeadCount:LogError(string.format(L["error.sort.name"], HeadCount.TITLE))
				return true
			end
		end)
	end
end

-- Sorts a player list by start time
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by start time.
function Raid:sortPlayerListByStartTime(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aBeginTime = a:retrieveStartingTime()
			local bBeginTime = b:retrieveStartingTime()

			if aBeginTime and bBeginTime then
				-- both a and b have valid end times
				local aSeconds = aBeginTime.epoc
				local bSeconds = bBeginTime.epoc

				if aSeconds ~= bSeconds then
					-- times are not equal
					if isDescending then
						return aSeconds < bSeconds
					else
						return aSeconds > bSeconds
					end
				else
					-- times are equivalent, sort by secondary value (player name)
					return a.name < b.name
				end
			else
				if aBeginTime then
					-- a has an end time, b does not
					if isDescending then
						return true
					else
						return false
					end
				elseif bBeginTime then
					-- b has an end time, a does not
					if isDescending then
						return false
					else
						return true
					end
				else
					-- neither a nor b have an end time, sort by secondary value (player name)
					return a.name < b.name
				end
			end
		end)
	end
end

-- Sorts a player list by end time
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by end time.
function Raid:sortPlayerListByEndTime(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aEndTime = a:retrieveEndingTime()
			local bEndTime = b:retrieveEndingTime()
			if type(aEndTime) == "number" then
				local t = aEndtime
				aEndTime = { epoc = t}
			end
			if  type(bEndTime) == "number" then
				local t = aEndtime
				bEndTime = { epoc = t}
			end
			if aEndTime and bEndTime then
				-- both a and b have valid end times
				local aSeconds = aEndTime.epoc or 0
				local bSeconds = bEndTime.epoc or 0

				if aSeconds ~= bSeconds then
					-- times are not equal
					if (isDescending) then
						return aSeconds < bSeconds
					else
						return aSeconds > bSeconds
					end
				else
					-- times are equivalent, sort by secondary value (player name)
					return a.name < b.name
				end
			else
				if aEndTime then
					-- a has an end time, b does not
					if isDescending then
						return true
					else
						return false
					end
				elseif bEndTime then
					-- b has an end time, a does not
					if isDescending then
						return false
					else
						return true
					end
				else
					-- neither a nor b have an end time, sort by secondary value (player name)
					return a.name < b.name
				end
			end
		end)
	end
end

-- Sorts a player list by total time
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by total time.
function Raid:sortPlayerListByTotalTime(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aTotalTime = a:getTotalTime()
			local bTotalTime = b:getTotalTime()

			if aTotalTime ~= bTotalTime then
				if isDescending then
					return aTotalTime < bTotalTime
				else
					return aTotalTime > bTotalTime
				end
			else
				return a.name < b.name
			end
		end)
	end
end

-- Sorts a player list by start time
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by start time.
function Raid:sortPlayerListByWaitlistActivityTime(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aWaitlistActivityTime = a.waitlistActivityTime
			local bWaitlistActivityTime = b.waitlistActivityTime

			if aWaitlistActivityTime and bWaitlistActivityTime then
				local aSeconds = aWaitlistActivityTime.epoc
				local bSeconds = bWaitlistActivityTime.epoc

				if aSeconds ~= bSeconds then
					-- times are not equal
					if isDescending then
						return aSeconds < bSeconds
					else
						return aSeconds > bSeconds
					end
				else
					-- times are equivalent, sory by secondary value (player name)
					return a.name < b.name
				end
			else
				if aWaitlistActivityTime then
					-- a has a waitlist activity time, b does not
					if isDescending then
						return true
					else
						return false
					end
				elseif bWaitlistActivityTime then
					-- b has a wait list activity time, a does not
					if isDescending then
						return false
					else
						return true
					end
				else
					-- neither a nor b have a waitlist activity time, sort by secondary value (player name)
					return a.name < b.name
				end
			end
		end)
	end
end

-- Sorts a player list by start time
-- @param orderedPlayerList The ordered player list.
-- @param isDescending The sort order.
-- @return table Returns the sorted player list by start time.
function Raid:sortPlayerListByWaitlistNote(orderedPlayerList, isDescending)
	if orderedPlayerList then
		table.sort(orderedPlayerList, function(a, b)
			local aWaitlistNote = a.waitlistNote
			local bWaitlistNote = b.waitlistNote

			if aWaitlistNote and bWaitlistNote then
				if aWaitlistNote ~= bWaitlistNote then
					-- notes are not equal
					if isDescending then
						return aWaitlistNote < bWaitlistNote
					else
						return aWaitlistNote > bWaitlistNote
					end
				else
					-- note are equivalent, sory by secondary value (player name)
					return a.name < b.name
				end
			else
				if aWaitlistNote then
					-- a has a waitlist note, b does not
					if isDescending then
						return true
					else
						return false
					end
				elseif bWaitlistNote then
					-- b has a wait list note, a does not
					if isDescending then
						return false
					else
						return true
					end
				else
					-- neither a nor b have a waitlist note, sort by secondary value (player name)
					return a.name < b.name
				end
			end
		end)
	end
end

-- Gets the number of players in the player list.
-- @return number Returns the number of players in the player list.
function Raid:getNumberOfWaitlistPlayers()
	local numberOfWaitlistPlayers = 0

	for k,v in pairs(self.playerList) do
		for i,j in pairs(self.playerList[k]) do
			if j.isWaitlisted then
				numberOfWaitlistPlayers = numberOfWaitlistPlayers + 1
			end
		end
	end

	return numberOfWaitlistPlayers
end

-- Adds a loot to the loot list
-- @param loot The loot to add to the loot list
function Raid:addLoot(loot)
	if loot then
		table.insert(self.lootList, loot)
	end
end

-- Removes a loot from the loot list
-- @param lootId The loot id.
function Raid:removeLoot(lootId)
	if self.lootList[lootId] then
		-- the loot exists
		table.remove(self.lootList, lootId)	-- remove the loot
	end
end

-- Retrieves a piece of loot by its loot id.
-- @param lootId The loot id
-- @return object A loot object or nil if the given loot does not exist
function Raid:retrieveLoot(lootId)
	return self.lootList[lootId]
end

-- Gets the loot list for a specific player.
-- @param playerName The player name.
-- @return table Returns the loot list for a specific player.
function Raid:getLootListByPlayer(playerName)
	local lootListByPlayer = { }

	if playerName then
		for k,v in ipairs(self.lootList) do
			if playerName == v.player then
				table.insert(lootListByPlayer, v)
			end
		end
	end

	return lootListByPlayer
end

-- Gets the begin time for the raid.
-- @return object Returns the raid start time or nil if none exists
function Raid:retrieveStartingTime()
	return self.timeList[1].start
end

-- Gets the end time for the raid.
-- @return number Returns the raid end time or nil if none exists
function Raid:retrieveEndingTime()
	local numberOfTimes = #self.timeList
	return self.timeList[numberOfTimes].stop
end

-- Finalize a raid
-- @param currentTime The finalization time.
function Raid:finalize(currentTime)
	if not self.isFinalized then
		-- finalize the player list
		for k,v in pairs(self.playerList["raidgroup"]) do
			self.playerList["raidgroup"][k]:finalize(currentTime)
		end

		for k,v in pairs(self.playerList["nogroup"]) do
			self.playerList["nogroup"][k]:finalize(currentTime)
		end
		-- Finalize the raid time
		local numberOfRaidTimePairs = #self.timeList
		local lastBeginTime = self.timeList[numberOfRaidTimePairs].start
		local lastEndTime = self.timeList[numberOfRaidTimePairs].stop
		if not lastEndTime or lastEndTime == 0 then
			-- ending raid time is not yet set
			self.timeList[numberOfRaidTimePairs].stop = currentTime
		end
		local timeDifference = HeadCount:computeTimeDifference(currentTime, lastBeginTime)
		self.raidTime = self.raidTime + timeDifference
		-- Add to the total raid time
		self.lastActivityTime = currentTime
		self.isFinalized = true
	end
end

-- Move a player to the raid group player list.
-- @param player The player
function Raid:moveToRaidGroup(player)
	if player then
		local playerName = player.name
		local isPresent, groupName = self:isPlayerPresent(playerName)
		if isPresent and "nogroup" == groupName then
			-- player is present
			-- player is in no group, add to the raid group
			self.playerList[groupName][playerName] = nil
			self.playerList["raidgroup"][playerName] = player
		else
			-- player is not present, add him
			self:addPlayer(player)
		end
	end
end

-- Move a player to the no group player list.
-- @param player The player
function Raid:moveToNoGroup(player)
	if player then
		local playerName = player.name
		local isPresent, groupName = self:isPlayerPresent(playerName)
		if isPresent and "raidgroup" == groupName then
			self.playerList[groupName][playerName] = nil
			self.playerList["nogroup"][playerName] = player
		else
			-- player is not present
			self:addPlayer(player)
		end
	end
end

-- Process players who may have left the raid.
-- @param numberOfRaidMembers The number of members currently in the actual raid.
-- @param trackedPlayerList The list of members currently in the actual raid.
-- @param activityTime The activity time.
function Raid:processMissingPlayers(numberOfRaidMembers, trackedPlayerList, activityTime)
	for k,v in pairs(self.playerList) do
		for i,j in pairs(self.playerList[k]) do
			local playerName = j.name
			if not trackedPlayerList[playerName] and not j.isFinalized then
				-- player is not currently in the actual raid and they haven't been finalized
				if (j:finalize(activityTime)) then
					HeadCount:LogDebug(string.format(L["debug.raid.update.modifymember.leave"], HeadCount.TITLE, HeadCount.VERSION, playerName))
				end
			end
		end
	end
end

-- Processes a begin time
-- @param activityTime The activity time.
-- @param note The note
function Raid:processBeginTime(activityTime, pnote)
	if activityTime then
		local numberOfRaidTimePairs = # self.timeList	-- get the total number of current time pairs
		local lastBeginTime = self.timeList[numberOfRaidTimePairs].start
		local lastEndTime = self.timeList[numberOfRaidTimePairs].stop

		if lastBeginTime and lastEndTime then
			-- both time values exist for the most recent pair, start a new pair
			local timePair = HeadCount:NewInstance("TimePair",{start = activityTime, note = pnote, stop = nil})
			table.insert(self.timeList, timePair)
		elseif not lastBeginTime and not lastEndTime then
			-- most recent pair contains no time values
			self.timeList[numberOfRaidTimePairs].start = activityTime
			self.timeList[numberOfRaidTimePairs].note = note
		end
	end
end

-- Processes an end time
-- @param activityTime The activity time.
function Raid:processEndTime(activityTime)
	if activityTime then
		local numberOfRaidTimePairs = # self.timeList	-- get the total number of current time pairs
		local lastBeginTime = self.timeList[numberOfRaidTimePairs].start
		local lastEndTime = self.timeList[numberOfRaidTimePairs].stop
		if lastBeginTime and not lastEndTime then
			-- the most recent beginning time exists
			-- no recent end time exists, add it
			self.timeList[numberOfRaidTimePairs].stop = activityTime
		end
	end
end

-- Recover from a fault.
-- @param activityTime The activity time.
function Raid:recover(activityTime)
	for k,v in pairs(self.playerList) do
		for i,j in pairs(self.playerList[k]) do
			local player = self.playerList[k][i]
			player:recover(activityTime, self.lastActivityTime)
		end
	end
	local timeIndex = #self.timeList
	local lastBeginTime = self.timeList[timeIndex].start
	local lastEndTime = self.timeList[timeIndex].stop
	if lastBeginTime and not lastEndTime then
		local timeDifference = HeadCount:computeTimeDifference(self.lastActivityTime, lastBeginTime)
		self.raidTime = self.raidTime + timeDifference
		self:processEndTime(self.lastActivityTime)
	end
	self:processBeginTime(activityTime, L["Recovery"])
	self.lastActivityTime = activityTime
	self.isFinalized = false
end


-- Add the Class Definition to Registry
HeadCount:DefineClass("Raid",Fields,Raid)